﻿
using System;
using System.Collections.Generic;

namespace TapLibrary
{
	// ChannelLFD.com
	// https://ChannelLFD.com/random-number-generator/
	//
	class TapRand
	{
		#region Variables

		private const UInt32 BEST_4_TAP_PATTERN = 0x801C0000;

		private UInt32[] _pinArray;
		private UInt32 _currentRandomValue;

		// To calculate the sequence length
		//private bool _isSequenceLengthKnown;
		private UInt32 _startValue;
		private UInt32 _sequenceCounter;
		private UInt32 _sequenceLength;

		#endregion

		#region Constructor

		public TapRand() 
		       : this(BEST_4_TAP_PATTERN, 1)
		{
		}

		public TapRand(UInt32 seedValue) 
		       : this(BEST_4_TAP_PATTERN, seedValue)
		{
		}

		public TapRand(UInt32 tapPattern, UInt32 seedValue)
		{
			// Set the start value
			_currentRandomValue = seedValue;

			// Set the tap pattern
			_pinArray = UintToPinlist(tapPattern);

			// Get a valid start pattern
			_startValue = Perform32Steps();
			// Clear the count we got from doing that
			_sequenceCounter = 0;
		}

		#endregion

		//----public-------------------

		#region NextRaw

		/// <summary>
		/// Returns null when interrupted, otherwise the results of the current test
		/// </summary>
		/// 
		public UInt32 NextRaw
		{
			get { return Perform32Steps(); }
		}

		#endregion

		#region NextFraction

		/// <summary>
		/// Returns a fraction in the range 0 to 0.999999999
		/// </summary>
		/// 
		public double NextFraction
		{
			get
			{
				const double TARGET_MAX_32BIT_FRACTION = 0.999999999;
				const double MAX_32BIT_RNG_VALUE = UInt32.MaxValue - 1.0;

				double currentValue = Perform32Steps();
				double scaledFraction = TARGET_MAX_32BIT_FRACTION * ((currentValue - 1.0) / MAX_32BIT_RNG_VALUE);

				return scaledFraction;
			}
		}

		#endregion

		#region IsSequenceComplete

		/// <summary>
		/// Returns sequence length when known, otherwise zero
		/// </summary>
		/// 
		public UInt32 SequenceLength
		{
			get
			{
				return _sequenceLength;
			}
		}

		#endregion

		//----private------------------

		#region Perform32Steps()

		/// <summary>
		/// Shift the result 32 times
		/// </summary>
		/// 
		private UInt32 Perform32Steps()
		{
			UInt32 currentVal = _currentRandomValue;
			UInt32 nextVal = 0;

			for (int i = 0; i < 32; i++)
			{
				nextVal = (currentVal << 1);

				foreach (UInt32 t in _pinArray)
					if ((currentVal & t) != 0)
						nextVal ^= 1;

				currentVal = nextVal;
			}

			_currentRandomValue = nextVal;

			if ((_startValue == _currentRandomValue) && (_sequenceLength == 0))
				_sequenceLength = _sequenceCounter;
			else
				_sequenceCounter++;

			return nextVal;
		}

		#endregion

		#region PinlistToUnit()

		// Converts a list of tap pins to a uint
		//
		private static UInt32 PinlistToUnit(UInt32[] pinList)
		{
			UInt32 tapPattern = 0;

			// Add in each pin in turn
			foreach (UInt32 t in pinList)
				tapPattern |= t;

			return tapPattern;
		}

		#endregion

		#region UintToPinlist()

		// Converts uint to a list of tap pins
		//
		private static UInt32[] UintToPinlist(UInt32 tapPattern)
		{
			List<UInt32> pinlist = new List<UInt32>();

			for (UInt32 i = 1; i != 0; i <<= 1)
				if ((tapPattern & i) != 0)
					pinlist.Add(i);

			return pinlist.ToArray();
		}

		#endregion
	}
}

